"""
The core module of the Rembus library that includes implementations for the
Router and Twin concepts.
"""
import asyncio
import base64
import logging
import os
import time
from typing import Callable, Any, Optional, List
import ssl
from urllib.parse import urlparse
import uuid
import async_timeout
from websockets.asyncio.server import serve
import cbor2
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding, rsa, ec
import websockets
import rembus.protocol as rp
import rembus.settings as rs
from . import __version__

logger = logging.getLogger(__name__)


def domain(s: str) -> str:
    """Return the domain part from the string.

    If no domain is found, return the root domain ".".
    """
    dot_index = s.find('.')
    if dot_index != -1:
        return s[dot_index + 1:]
    else:
        return "."


def randname() -> str:
    """Return a random name for a component."""
    return str(uuid.uuid4())


def bytes_to_b64(val: bytes, enc: int):
    """Base 64 encodeing for JSON-RPC transport"""
    if enc == rp.JSON:
        return base64.b64encode(val).decode("utf-8")
    return val


async def get_response(obj: Any) -> Any:
    """Return the response of the object."""
    if asyncio.iscoroutine(obj):
        return await obj
    else:
        return obj


class FutureResponse:
    """
    Encapsulate a future response for a request.
    """

    def __init__(self, data: Any = None):
        self.future = asyncio.get_running_loop().create_future()
        self.data = data


def getargs(data):
    """
    Return arguments list from the data.
    """
    if isinstance(data, list):
        return data
    else:
        return [data]


class RbURL:
    """
    A class to parse and manage Rembus URLs.
    It supports the 'repl' scheme and the standard 'ws'/'wss' schemes.
    """

    def __init__(self, url: str | None = None) -> None:
        default_url = os.getenv('REMBUS_BASE_URL', "ws://127.0.0.1:8000")
        baseurl = urlparse(default_url)
        uri = urlparse(url)

        if uri.scheme == "repl":
            self.protocol = uri.scheme
            self.hostname = ''
            self.port = 0
            self.hasname = False
            self.id = 'repl'
        else:
            if isinstance(uri.path, str) and uri.path:
                self.hasname = True
                self.id = uri.path[1:] if uri.path.startswith(
                    "/") else uri.path
            else:
                self.hasname = False
                self.id = randname()

            if uri.scheme:
                self.protocol = uri.scheme
            else:
                self.protocol = baseurl.scheme

            if uri.hostname:
                self.hostname = uri.hostname
            else:
                self.hostname = baseurl.hostname

            if uri.port:
                self.port = uri.port
            else:
                self.port = baseurl.port

    def __repr__(self):
        return f"{self.protocol}://{self.hostname}:{self.port}/{self.id}"

    def isrepl(self):
        """Check if the URL is a REPL."""
        return self.protocol == "repl"

    def connection_url(self):
        """Return the URL string."""
        if self.hasname:
            return f"{self.protocol}://{self.hostname}:{self.port}/{self.id}"
        else:
            return f"{self.protocol}://{self.hostname}:{self.port}"

    @property
    def netlink(self):
        """Return the remote connection endpoint"""
        return f"{self.protocol}://{self.hostname}:{self.port}"

    @property
    def twkey(self):
        """Return the twin key"""
        return f"{self.id}@{self.netlink}"


async def shutdown_message(obj) -> None:
    """Log a shutdown message for the given object."""
    logger.debug("[%s] sending shutdown message", obj)
    await obj.inbox.put("shutdown")


class Supervised:
    """
    A superclass that provides task supervision and auto-restarting for
    a designated task.
    Subclasses must implement the '_task_impl' coroutine.
    """

    def __init__(self):
        self._task: Optional[asyncio.Task[None]] = None
        self._supervisor_task: Optional[asyncio.Task[None]] = None
        self._should_run = True  # Flag to control supervisor loop

    async def _shutdown(self) -> None:
        """Override in subclasses for custom shutdown logic."""

    async def _task_impl(self) -> None:
        """Override in subclasses for supervised task impl."""

    async def _supervisor(self) -> None:
        """
        Supervises the _task_impl, restarting if it exits
        unexpectedly or due to an exception.
        """
        while self._should_run:
            logger.debug("[%s] starting supervised task", self)
            self._task = asyncio.create_task(self._task_impl())
            try:
                await self._task
            except asyncio.CancelledError:
                logger.debug("[%s] task cancelled, exiting", self)
                self._should_run = False  # Ensure supervisor also stops
                break
            except Exception as e:  # pylint: disable=broad-exception-caught
                logger.error("[%s] error: %s (restarting)", self, e)
                logging.exception("traceback for task error:")
                if self._should_run:
                    await asyncio.sleep(0.5)

    def start(self) -> None:
        """Starts the supervisor task."""
        self._should_run = True
        self._supervisor_task = asyncio.create_task(self._supervisor())

    async def shutdown(self) -> None:
        """Gracefully stops the supervised worker and its supervisor."""
        logger.debug("[%s] shutting down (should_run: %s)",
                     self, self._should_run)
        if self._should_run:
            self._should_run = False

            await self._shutdown()

            if self._task and not self._task.done():
                await shutdown_message(self)
                await self._task

            if self._supervisor_task and not self._supervisor_task.done():
                self._supervisor_task.cancel()
                try:
                    await self._supervisor_task
                except asyncio.CancelledError:
                    pass
            logger.debug("[%s] shutdown complete", self)


async def init_router(router_name, uid, port, secure):
    """Start the router"""
    router = Router(router_name)
    logger.debug("component %s created, port: %s", uid.id, port)
    # start a websocket server
    if port:
        router.serve_task = asyncio.create_task(router.serve_ws(port, secure))
        done, _ = await asyncio.wait([router.serve_task], timeout=0.1)
        if router.serve_task in done:
            try:
                await router.serve_task
            except Exception as e:
                router.serve_task = None
                logger.error("[%s] start failed: %s", router, e)
                await router.shutdown()
                raise
    else:
        router.serve_task = None

    return router


class Router(Supervised):
    """
    A Router is a central component that manages connections and interactions
    between Rembus components(Twins).
    """

    def __init__(self, name: str):
        super().__init__()
        self.id = name
        self.id_twin: dict[str, Twin] = {}
        self.handler: dict[str, Callable[..., Any]] = {}
        self.inbox: asyncio.Queue[Any] = asyncio.Queue()
        self.shared: Any = None
        self.serve_task: Optional[asyncio.Task[None]] = None
        self.server_instance = None  # To store the server object
        self._shutdown_event = asyncio.Event()  # For controlled shutdown
        self.config = rs.Config(name)
        self.owners = rs.load_tenants(self)
        self.start_ts = time.time()
        self._builtins()
        self.start()

    def __str__(self):
        return f"{self.id}"

    def __repr__(self):
        return f"{self.id}: {self.id_twin}"

    def isconnected(self, rid: str) -> bool:
        """Check if a component with the given rid is connected."""
        for tk in self.id_twin:
            if tk.startswith(rid+'@'):
                return True
        return False

    def uptime(self) -> str:
        """Return the uptime of the router."""
        return f"up for {int(time.time() - self.start_ts)} seconds"

    def _builtins(self):
        self.handler["rid"] = lambda *_: self.id
        self.handler["version"] = lambda *_: __version__
        self.handler["uptime"] = lambda *_: self.uptime()

    async def init_twin(self, uid: RbURL, enc: int, isserver: bool):
        """Create and start a Twin"""
        cmp = Twin(uid, self, not isserver, enc)
        if not uid.isrepl():
            self.id_twin[uid.twkey] = cmp

        try:
            # if not isserver:
            if not cmp.isrepl():
                if self.config.start_anyway:
                    await cmp.inbox.put("reconnect")
                else:
                    await cmp.connect()
        except Exception:
            await cmp.close()
            raise
        return cmp

    async def _shutdown(self):
        """Cleanup logic when shutting down the router."""
        logger.debug("[%s] router shutdown", self)
        if self.server_instance:
            self.server_instance.close()
            self._shutdown_event.set()
            await self.server_instance.wait_closed()

    async def _pubsub_msg(self, msg: rp.PubSubMsg):
        twin = msg.twin
        qos = msg.flags & rp.QOS2
        if qos > rp.QOS0 and msg.id:
            if twin.socket:
                await twin.send(rp.AckMsg(id=msg.id))

            if qos == rp.QOS2:
                if msg.id in twin.ackdf:
                    # Already received, skip the message.
                    return
                else:
                    # Save the message id to guarantee exactly one delivery.
                    twin.ackdf[msg.id] = int(time.time())

        data = rp.tag2df(msg.data)
        try:
            if msg.topic in self.handler:
                await self.evaluate(self, msg.topic, data)

            for t in self.id_twin.values():
                if t != twin:
                    # Do not send back to publisher.
                    await t.send(msg)

        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.warning("[%s] error in method invocation: %s", self, e)

        return

    async def send_message(self, msg):
        """Send message to remote node using a twin from the pool of twins"""
        for t in self.id_twin.values():
            if t.isopen():
                try:
                    sts = rp.STS_OK
                    data = await t.msg_send_wait(msg)
                except Exception as e:  # pylint:disable=broad-exception-caught
                    sts = rp.STS_ERROR
                    data = f"{e}"
                fut = await msg.twin.future_request(msg.id)
                if fut:
                    if sts == rp.STS_OK:
                        fut.future.set_result(rp.tag2df(data))
                    else:
                        fut.future.set_exception(
                            rp.RembusError(rp.STS_ERROR, data))
                break

    async def _rpcreq_msg(self, msg: rp.RpcReqMsg):
        """Handle an RPC request."""
        data = rp.tag2df(msg.data)

        if msg.twin.isrepl():
            await self.send_message(msg)
        elif msg.topic in self.handler:
            status = rp.STS_OK
            try:
                output = await self.evaluate(self, msg.topic, data)
            except Exception as e:  # pylint: disable=broad-exception-caught
                status = rp.STS_METHOD_EXCEPTION
                output = f"{e}"
                logger.debug("exception: %s", e)
            outmsg = rp.ResMsg(
                id=msg.id, status=status, data=rp.df2tag(output)
            )
            await msg.twin.send(outmsg)
        else:
            outmsg = rp.ResMsg(
                id=msg.id, status=rp.STS_METHOD_NOT_FOUND, data=msg.topic
            )
            await msg.twin.send(outmsg)
        return

    async def _handle_identity(self, msg: rp.IdentityMsg) -> None:
        twin_id = msg.cid
        if self.isconnected(twin_id):
            logger.warning("[%s] node with id [%s] is already connected", self, twin_id)
            await msg.twin.close()
        else:
            logger.debug("[%s] identity: %s", self, msg.cid)
            await self._auth_identity(msg)

    async def _handle_attestation(self, msg: rp.AttestationMsg) -> None:
        sts = self._verify_signature(msg)
        await msg.twin.response(sts, msg)

    async def _handle_admin(self, msg: rp.AdminMsg) -> None:
        logger.debug("[%s] admin: %s", self, msg)
        if msg.twin.isrepl():
            await self.send_message(msg)
        else:
            await msg.twin.response(rp.STS_OK, msg)

    async def _task_impl(self) -> None:
        logger.debug("[%s] router started", self)
        while True:
            msg = await self.inbox.get()
            match msg:
                case "shutdown":
                    logger.debug("[%s] router shutting down", self)
                    break
                case rp.PubSubMsg():
                    await self._pubsub_msg(msg)
                case rp.RpcReqMsg():
                    await self._rpcreq_msg(msg)
                case rp.IdentityMsg():
                    await self._handle_identity(msg)
                case rp.AttestationMsg():
                    await self._handle_attestation(msg)
                case rp.AdminMsg():
                    await self._handle_admin(msg)
                case rp.RegisterMsg():
                    await self._register_node(msg)
                case rp.UnregisterMsg():
                    await self._unregister_node(msg)


    async def evaluate(self, twin, topic: str, data: Any) -> Any:
        """Invoke the handler associate with the message topic."""
        if self.shared is not None:
            output = await get_response(
                self.handler[topic](self.shared, twin, *getargs(data))
            )
        else:
            output = await get_response(self.handler[topic](*getargs(data)))

        return output

    async def _client_receiver(self, ws):
        """Receive messages from the client component."""
        url = RbURL()
        twin = Twin(url, self, False)
        self.id_twin[url.twkey] = twin
        twin.socket = ws
        await twin.twin_receiver()

    async def serve_ws(self, port: int, issecure: bool = False):
        """Start a WebSocket server to handle incoming connections."""
        ssl_context = None
        if issecure:
            trust_store = rs.keystore_dir()
            cert_path = os.path.join(trust_store, "rembus.crt")
            key_path = os.path.join(trust_store, "rembus.key")
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            if not os.path.isfile(cert_path) or not os.path.isfile(key_path):
                raise RuntimeError(f"SSL secrets not found in {trust_store}")

            ssl_context.load_cert_chain(cert_path, keyfile=key_path)

        async with serve(
                self._client_receiver,
                "0.0.0.0",
                port,
                ssl=ssl_context,
                ping_interval=self.config.ws_ping_interval,) as server:
            self.server_instance = server
            await self._shutdown_event.wait()
            # await server.serve_forever()

    def _needs_auth(self, cid: str):
        """Check if the component needs authentication."""
        try:
            rs.key_file(self.id, cid)
            return True
        except FileNotFoundError:
            return False

    def _update_twin(self, twin, identity):
        logger.debug("[%s] setting name: [%s]", twin, identity)
        self.id_twin.pop(twin.twkey, twin)
        twin.rid = identity
        self.id_twin[twin.twkey] = twin

    def _verify_signature(self, msg: rp.AttestationMsg):
        """Verify the signature of the attestation message."""
        twin = msg.twin
        cid = msg.cid
        fn = twin.handler.pop("challenge")
        challenge = fn(twin)
        plain = cbor2.dumps([challenge, msg.cid])
        try:
            if isinstance(msg.signature, str):
                signature = base64.b64decode(msg.signature)
            else:
                signature = msg.signature

            pubkey = rp.load_public_key(self, cid)
            if isinstance(pubkey, rsa.RSAPublicKey):
                pubkey.verify(signature, plain,
                              padding.PKCS1v15(), hashes.SHA256())
            elif isinstance(pubkey, ec.EllipticCurvePublicKey):
                pubkey.verify(signature, plain, ec.ECDSA(hashes.SHA256()))

            self._update_twin(twin, msg.cid)
            return rp.STS_OK
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error("verification failed: %s (%s)", e, type(e))
            return rp.STS_ERROR

    def _challenge(self, msg: rp.IdentityMsg):
        """Generate a challenge for the identity authentication."""
        twin = msg.twin
        challenge_val = os.urandom(4)
        twin.handler["challenge"] = lambda twin: challenge_val
        return rp.ResMsg(
            id=msg.id,
            status=rp.STS_CHALLENGE,
            data=bytes_to_b64(challenge_val, twin.enc)
        )

    async def _auth_identity(self, msg: rp.IdentityMsg):
        """Authenticate the identity of the component."""
        twin = msg.twin
        identity = msg.cid

        if self._needs_auth(identity):
            # component is provisioned, send the challenge
            response = self._challenge(msg)
        else:
            self._update_twin(twin, identity)
            response = rp.ResMsg(id=msg.id, status=rp.STS_OK)

        await twin.send(response)

    def _get_token(self, tenant, secret: str):
        """Get the token embedded into the message id."""
        pin = self.owners.get(tenant)
        if secret != pin:
            logger.info("tenant %s: invalid token %s", tenant, secret)
            return None
        else:
            logger.debug("tenant %s: token is valid", tenant)
            return secret

    async def _register_node(self, msg: rp.RegisterMsg):
        """Provision a new node."""
        sts = rp.STS_ERROR
        reason = None
        token = self._get_token(domain(msg.cid), msg.pin)
        try:
            if token is None:
                reason = "wrong tenant/pin"
            elif rp.isregistered(self.id, msg.cid):
                sts = rp.STS_NAME_ALREADY_TAKEN
                reason = f"[{msg.cid}] not available"
            else:
                kdir = rs.keys_dir(self.id)
                os.makedirs(kdir, exist_ok=True)
                rp.save_pubkey(self.id, msg.cid, msg.pubkey, msg.type)
                sts = rp.STS_OK
                logger.debug("cid %s registered", msg.cid)
        finally:
            await msg.twin.response(sts, msg, reason)

    async def _unregister_node(self, msg: rp.UnregisterMsg):
        """Unprovisions the component."""
        sts = rp.STS_ERROR
        reason = None
        try:
            cid = msg.twin.rid
            rp.remove_pubkey(self, cid)
            sts = rp.STS_OK
        finally:
            await msg.twin.response(sts, msg, reason)


class Twin(Supervised):
    """
A Twin represents a Rembus component, either as a client or server.
It handles the connection, message sending and receiving, and provides methods
for RPC, pub/sub, and other commands interactions.
    """

    def __init__(
            self,
            uid: RbURL,
            router: Router,
            isclient: bool = True,
            enc: int = rp.CBOR):
        super().__init__()
        self.isclient = isclient
        self.enc = enc
        self._router = router
        self.socket: websockets.ClientConnection | None = None
        self.receiver = None
        self.uid = uid
        self.inbox: asyncio.Queue[str] = asyncio.Queue()
        self.handler: dict[str, Callable[..., Any]] = {}
        self.outreq: dict[int, FutureResponse] = {}
        self.reconnect_task: Optional[asyncio.Task[None]] = None
        self.ackdf: dict[int, int] = {}  # msgid => ts
        self.handler["phase"] = lambda: "CLOSED"
        self.start()

    def __str__(self):
        return f"{self.uid.id}"

    def __repr__(self):
        return self.uid.id

    @property
    def rid(self):
        """Return the unique id of the rembus component."""
        return self.uid.id

    @rid.setter
    def rid(self, rid: str):
        self.uid.id = rid

    @property
    def twkey(self):
        """Return the twin key"""
        return f"{self.uid.twkey}"

    @property
    def router(self):
        """Return the router associated with this twin."""
        return self._router

    def isrepl(self) -> bool:
        """Check if twin is a REPL"""
        return self.uid.protocol == "repl"

    def isopen(self) -> bool:
        """Check if the connection is open."""
        if self.isrepl():
            return any([t.isopen() for t in self.router.id_twin.values()])
        else:
            return (self.socket is not None and
                    self.socket.state == websockets.State.OPEN)

    async def response(self, status: int, msg: Any, data: Any = None):
        """Send a response to the client."""
        outmsg: Any = rp.ResMsg(id=msg.id, status=status, data=data)
        await self.send(outmsg)

    def inject(self, data: Any):
        """Initialize the context object."""
        self.router.shared = data

    async def _reconnect(self):
        logger.debug("[%s]: reconnecting ...", self)
        while True:
            try:
                await self.connect()
                await self.reactive()
                self.reconnect_task = None
                break
            except Exception as e:  # pylint: disable=broad-exception-caught
                logger.info("[%s] reconnect: %s", self, e)
                await asyncio.sleep(2)

    async def _shutdown(self):
        """Cleanup logic when shutting down the twin."""
        logger.debug("[%s] twin shutdown", self)

        if self.isclient or self.uid.isrepl():
            await self._shutdown_router()

        await self._close_socket()
        await self._cancel_task("receiver")
        await self._cancel_task("reconnect_task")

        if self.uid.isrepl():
            await self._shutdown_twins()


    async def _shutdown_router(self):
        if self.router:
            await self.router.shutdown()


    async def _close_socket(self):
        if self.socket:
            await self.socket.close()
            self.socket = None


    async def _cancel_task(self, name: str):
        task = getattr(self, name, None)
        if task:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass
            setattr(self, name, None)


    async def _shutdown_twins(self):
        # Close the twins (or the components of the pool)
        for t in list(self.router.id_twin.values()):
            t.handler["phase"] = lambda: "CLOSED"
            if t.socket is not None:
                await t.shutdown()


    async def _task_impl(self):
        logger.debug("[%s] task started", self)
        while True:
            msg: str = await self.inbox.get()
            logger.debug("[%s] twin_task: %s", self, msg)
            if msg == "reconnect":
                if not self.reconnect_task:
                    self.reconnect_task = asyncio.create_task(
                        self._reconnect())
            elif msg == "shutdown":
                break

    async def twin_receiver(self):
        """Receive messages from the WebSocket connection."""
        logger.debug("[%s] client is connected", self)
        try:
            while self.socket is not None:
                result: str | bytes = await self.socket.recv()
                if isinstance(result, str):
                    self.enc = rp.JSON
                    msg = rp.jsonrpc_parse(result)
                else:
                    self.enc = rp.CBOR
                    pkt: list[Any] = cbor2.loads(result)
                    msg = rp.cbor_parse(pkt)

                msg.twin = self
                await self._eval_input(msg)
        except (
            websockets.ConnectionClosedOK,
            websockets.ConnectionClosedError
        ) as e:
            logger.debug("connection closed: %s", e)
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.warning("[%s] error: %s", self, e)
        finally:
            if self.isclient and self.handler["phase"]() == "CONNECTED":
                logger.debug("[%s] twin_receiver done", self)
                await self.inbox.put("reconnect")
            else:
                self.router.id_twin.pop(self.twkey, None)
                await self.shutdown()

            self.receiver = None

    async def future_request(self, msgid: int):
        """Return the future associated with the message id `msgid`."""
        fut = self.outreq.pop(msgid, None)
        if fut is None:
            logger.warning("[%s] recv unknown msg id %s",
                           self, rp.tohex(rp.to_bytes(msgid)))
        elif fut.future.done():
            return None

        return fut

    async def _response_msg(self, msg: rp.ResMsg):
        fut = await self.future_request(msg.id)
        if fut:
            sts = msg.status
            if sts == rp.STS_OK:
                fut.future.set_result(rp.tag2df(msg.data))
            elif sts == rp.STS_CHALLENGE:
                fut.future.set_result(msg.data)
            else:
                fut.future.set_exception(rp.RembusError(sts, msg.data))

    async def _ack_msg(self, msg: rp.AckMsg):
        logger.debug("[%s] pubsub ack: %s", self, msg.id)
        fut = await self.future_request(msg.id)
        if fut:
            if fut.data:
                await self.send(rp.Ack2Msg(id=msg.id))

            fut.future.set_result(True)

    async def _ack2_msg(self, msg: rp.Ack2Msg):
        mid = msg.id
        if mid in self.ackdf:
            logger.debug("[%s] ack2: deleting pubsub ack %s", self, mid)
            del self.ackdf[mid]
        return

    async def _eval_input(self, msg: rp.RembusMsg):
        """
        Receive the incoming message and dispatch
        it to the appropriate handler.
        """

        if isinstance(msg, rp.ResMsg):
            await self._response_msg(msg)
        elif isinstance(msg, rp.AckMsg):
            await self._ack_msg(msg)
        elif isinstance(msg, rp.Ack2Msg):
            await self._ack2_msg(msg)
        else:
            self.router.inbox.put_nowait(msg)

    async def connect(self):
        """Connect to the broker."""
        broker_url = self.uid.connection_url()
        ssl_context = None
        if self.uid.protocol == "wss":
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            ca_crt = os.getenv("HTTP_CA_BUNDLE", rs.rembus_ca())
            if os.path.isfile(ca_crt):
                ssl_context.load_verify_locations(ca_crt)
            else:
                logger.warning("CA file not found: %s", ca_crt)

        self.socket = await websockets.connect(
            broker_url,
            ping_interval=self.router.config.ws_ping_interval,
            max_size=rp.WS_FRAME_MAXSIZE,
            ssl=ssl_context
        )
        self.handler["phase"] = lambda: "CONNECTING"
        self.receiver = asyncio.create_task(self.twin_receiver())

        if self.uid.hasname:
            try:
                await self._login()
            except Exception as e:
                await self.close()
                raise rp.RembusError(rp.STS_ERROR, "_login failed") from e

        self.handler["phase"] = lambda: "CONNECTED"
        return self

    async def send(self, msg: rp.RembusMsg):
        """Send a rembus message"""
        if self.isrepl() and self.isopen():
            await self.router.inbox.put(msg)
        elif self.socket is None:
            raise rp.RembusConnectionClosed()
        else:
            pkt = msg.to_payload(self.enc)
            await self._send(pkt)

    async def _send(self, payload: bytes | str) -> Any:
        if self.socket is not None:
            await self.socket.send(payload)

    async def _qos_send_wait(
            self,
            msgid: int,
            msg: rp.PubSubMsg) -> Any:
        """Send a pubsub message and wait for the ack."""
        await self.send(msg)
        futreq = FutureResponse(msg.flags & rp.QOS2 == rp.QOS2)
        self.outreq[msgid] = futreq
        try:
            async with async_timeout.timeout(
                self.router.config.request_timeout
            ):
                return await futreq.future
        except TimeoutError as e:
            raise rp.RembusTimeout() from e

    async def _send_wait(
            self,
            builder: Callable,
            data: Any = None) -> Any:
        """Send a message and wait for a response."""
        reqid = rp.msgid()
        req = builder(reqid)
        req.twin = self
        await self.send(req)
        futreq = FutureResponse(data)
        self.outreq[reqid] = futreq
        try:
            async with async_timeout.timeout(
                self.router.config.request_timeout
            ):
                return await futreq.future
        except TimeoutError as e:
            raise rp.RembusTimeout() from e

    async def msg_send_wait(
            self,
            msg: rp.RpcReqMsg) -> Any:
        """Send a message and wait for a response."""
        await self.send(msg)
        futreq = FutureResponse(msg.data)
        self.outreq[msg.id] = futreq
        try:
            async with async_timeout.timeout(
                self.router.config.request_timeout
            ):
                return await futreq.future
        except TimeoutError as e:
            raise rp.RembusTimeout() from e

    async def _login(self):
        """Connect in free mode or authenticate the provisioned component."""
        challenge = await self._send_wait(
            lambda id: rp.IdentityMsg(id=id, cid=self.uid.id)
        )
        if isinstance(challenge, bytes | str):
            if isinstance(challenge, str):
                challenge = base64.b64decode(challenge)

            plain = [bytes(challenge), self.uid.id]
            message = cbor2.dumps(plain)
            privatekey = rp.load_private_key(self.uid.id)
            if isinstance(privatekey, rsa.RSAPrivateKey):
                signature: bytes = privatekey.sign(
                    message, padding.PKCS1v15(), hashes.SHA256())
            elif isinstance(privatekey, ec.EllipticCurvePrivateKey):
                signature: bytes = privatekey.sign(
                    message, ec.ECDSA(hashes.SHA256()))

            await self._send_wait(
                lambda id: rp.AttestationMsg(
                    id=id,
                    cid=self.uid.id,
                    signature=bytes_to_b64(signature, self.enc)
                ))
        else:
            logger.debug("[%s]: free mode access", self)

    async def publish(self, topic: str, *data: Any, **kwargs):
        """Publish a message to the specified topic."""

        slot = kwargs.get("slot", None)
        qos = kwargs.get("qos", rp.QOS0)
        if qos == rp.QOS0:
            msg = rp.PubSubMsg(topic=topic, data=data, slot=slot)
            msg.twin = self
            if self.isrepl() and self.isopen():
                await self.router.inbox.put(msg)
            elif self.socket is None:
                raise rp.RembusConnectionClosed()
            else:
                await self.send(msg)
        else:
            await self._qos_publish(topic, data, qos, slot)

        return None

    async def put(self, topic: str, *args: Any, **kwargs):
        """Publish a message to the topic prefixed with component name."""
        await self.publish(self.rid + '/' + topic, *args, **kwargs)

    async def _qos_publish(
            self,
            topic: str,
            data: tuple,
            qos: rp.UInt8,  # type: ignore[valid-type]
            slot: int | None
    ):
        done = False
        max_retries = self.router.config.send_retries
        retries = 0
        if slot is None:
            reqid = rp.msgid()
        else:
            reqid = rp.msgid_slot(slot)

        while True:
            retries += 1
            try:
                done = await self._qos_send_wait(
                    reqid,
                    rp.PubSubMsg(
                        id=reqid, topic=topic, data=data, flags=qos, slot=slot
                    )
                )

                if done:
                    break
            except rp.RembusTimeout:
                if retries > max_retries:
                    raise rp.RembusTimeout() from None

    async def broker_setting(self, command: str, args: dict[str, Any]):
        """Send a broker configuration command."""
        data = {rp.COMMAND: command} | args
        return await self._send_wait(
            lambda id: rp.AdminMsg(id=id, topic=rp.BROKER_CONFIG, data=data)
        )

    async def setting(
            self, topic: str, command: str, args: dict[str, Any] | None = None
    ):
        """Send an admin command to the broker."""
        if self.socket:
            if args:
                data = {rp.COMMAND: command} | args
            else:
                data = {rp.COMMAND: command}

            return await self._send_wait(
                lambda id: rp.AdminMsg(id=id, topic=topic, data=data)
            )

    async def rpc(self, topic: str, *args: Any):
        """Send a RPC request."""
        data = rp.df2tag(args)
        return await self._send_wait(
            lambda id: rp.RpcReqMsg(id=id, topic=topic, data=data)
        )

    async def direct(self, target: str, topic: str, *args: Any):
        """Send a RPC request to a specific target."""
        data = rp.df2tag(args)
        return await self._send_wait(
            lambda id: rp.RpcReqMsg(
                id=id, topic=topic, target=target, data=data)
        )

    async def register(self, rid: str, pin: str, scheme: int = rp.SIG_RSA):
        """Provisions the component with rid identifier."""
        if scheme == rp.SIG_RSA:
            privkey = rp.rsa_private_key()
        else:
            privkey = rp.ecdsa_private_key()

        pubkey = rp.pem_public_key(privkey)
        if self.enc == rp.JSON:
            pubkey = base64.b64encode(pubkey).decode("utf-8")

        response = await self._send_wait(
            lambda id: rp.RegisterMsg(
                id=id, cid=rid, pin=pin, pubkey=pubkey, type=scheme)
        )

        logger.debug("cid %s registered", rid)
        rp.save_private_key(rid, privkey)
        return response

    async def unregister(self):
        """Unprovisions the component."""
        await self._send_wait(
            lambda id: rp.UnregisterMsg(id=id)
        )
        os.remove(os.path.join(rs.rembus_dir(), self.uid.id, ".secret"))
        return self

    async def reactive(self):
        """
        Set the component to receive published messages on subscribed topics.
        """
        if self.isclient:
            await self.broker_setting("reactive", {"status": True})
        return self

    async def unreactive(self):
        """
        Set the component to stop receiving published
        messages on subscribed topics.
        """
        await self.broker_setting("reactive", {"status": False})
        return self

    async def subscribe(
            self,
            fn: Callable[..., Any],
            retroactive: bool = False,
            topic: Optional[str] = None
    ):
        """
        Subscribe the function to the corresponding topic.
        """
        if topic is None:
            topic = fn.__name__

        await self.setting(
            topic, rp.ADD_INTEREST, {"retroactive": retroactive}
        )
        self.router.handler[topic] = fn
        return self

    async def unsubscribe(self, fn: Callable[..., Any]):
        """
        Unsubscribe the function from the corresponding topic.
        """
        topic = fn.__name__
        await self.setting(topic, rp.REMOVE_INTEREST)
        self.router.handler.pop(topic, None)
        return self

    async def expose(self, fn: Callable[..., Any]):
        """
        Expose the function as a remote procedure call(RPC) handler.
        """
        topic = fn.__name__
        self.router.handler[topic] = fn
        await self.setting(topic, rp.ADD_IMPL)

    async def unexpose(self, fn: Callable[..., Any]):
        """
        Unexpose the function as a remote procedure call(RPC) handler.
        """
        topic = fn.__name__
        self.router.handler.pop(topic, None)
        await self.setting(topic, rp.REMOVE_IMPL)

    async def close(self):
        """Close the connection and clean up resources."""
        self.handler["phase"] = lambda: "CLOSED"
        await self.shutdown()
        await asyncio.sleep(0)  # let loop drain tasks

    async def wait(self, timeout: float | None = None):
        """
        Start the twin event loop that wait for rembus messages.
        """
        if not self.isrepl():
            await self.reactive()
        if self._supervisor_task is not None:
            try:
                await asyncio.wait([self._supervisor_task], timeout=timeout)
            finally:
                await self.shutdown()


async def component(
    url: str | List[str] | None = None,
    name: str | None = None,
    port: int | None = None,
    secure: bool = False,
    enc: int = rp.CBOR
) -> Twin:
    """Return a Rembus component."""
    isserver = (port is not None) and (url is None)
    if isinstance(url, str):
        uid = RbURL(url)
    else:
        uid = RbURL("repl://")

    default_name = rs.DEFAULT_BROKER
    if uid.hasname:
        default_name = uid.id

    router_name = name if name else default_name
    router = await init_router(router_name, uid, port, secure)

    handle = await router.init_twin(uid, enc, isserver)
    if isinstance(url, list):
        for netlink in url:
            await router.init_twin(RbURL(netlink), enc, isserver)

    return handle



"""Settings for Rembus components."""
import json
import logging
import os
from platformdirs import user_config_dir

logger = logging.getLogger(__name__)

DEFAULT_BROKER = "broker"
TENANTS_FILE = "tenants.json"


class Config:
    """Configuration values to modify the behavior of Rembus."""

    def __init__(self, name: str):
        cfg = {}
        try:
            fn = os.path.join(rembus_dir(), name, "settings.json")
            if os.path.isfile(fn):
                with open(fn, 'r', encoding='utf-8') as f:
                    cfg = json.load(f)
        except json.decoder.JSONDecodeError as e:
            raise (RuntimeError(f"{fn}: {e}")) from e

        self.request_timeout = cfg.get("request_timeout", 5)
        self.ws_ping_interval = cfg.get("ws_ping_interval", None)
        self.start_anyway = cfg.get("start_anyway", False)
        self.send_retries = cfg.get("send_retries", 3)


def rembus_dir():
    """The root directory for all rembus components."""
    rdir = os.getenv("REMBUS_DIR")
    if not rdir:
        return user_config_dir("rembus", "Rembus")

    return rdir


def broker_dir(router_id):
    """The directory for rembus broker settings and secrets."""
    return os.path.join(rembus_dir(), router_id)


def keys_dir(router_id: str):
    """The directory for rembus public keys."""
    return os.path.join(broker_dir(router_id), "keys")


def keystore_dir():
    """The directory for rembus keystore."""
    return os.environ.get(
        "REMBUS_KEYSTORE", os.path.join(rembus_dir(), "keystore")
    )


def rembus_ca():
    """The CA bundle for rembus located at default location."""
    cadir = os.path.join(rembus_dir(), "ca")
    if os.path.isdir(cadir):
        files = os.listdir(cadir)
        if len(files) == 1:
            return os.path.join(cadir, files[0])

    raise RuntimeError("CA bundle not found")


def key_base(broker_name: str, cid: str) -> str:
    """The directory for the public keys."""
    return os.path.join(rembus_dir(), broker_name, "keys", cid)


def key_file(broker_name: str, cid: str):
    """The public key file for the given component."""
    basename = key_base(broker_name, cid)

    for keyformat in ["pem", "der"]:
        for keytype in ["rsa", "ecdsa"]:
            fn = f"{basename}.{keytype}.{keyformat}"
            logger.debug("looking for %s", fn)
            if os.path.isfile(fn):
                return fn

    raise FileNotFoundError(f"key file not found: {basename}")


def load_tenants(router):
    """Load the tenants settings."""
    fn = os.path.join(broker_dir(router.id), TENANTS_FILE)
    cfg = {}
    if os.path.isfile(fn):
        with open(fn, "r", encoding="utf-8") as f:
            cfg = json.load(f)
    return cfg



"""
Utility functions, messages and protocol constants related
to the Rembus protocol.
"""
import base64
import os
import logging
from typing import Any, List
import json
import cbor2
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa, ec
from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes
from cryptography.hazmat.backends import default_backend
from narwhals.typing import IntoFrame
import polars as pl
import pandas as pd
import pyarrow as pa
from pydantic import BaseModel, PrivateAttr, conint
import rembus.settings as rs

logger = logging.getLogger(__name__)

WS_FRAME_MAXSIZE = 60 * 1024 * 1024

SIG_RSA = 0x1
SIG_ECDSA = 0x2

CBOR = 0
JSON = 1

UInt8 = conint(ge=0, le=255)

QOS0 = UInt8(0x00)
QOS1 = UInt8(0x10)
QOS2 = UInt8(0x30)
SLOT_FLAG = UInt8(0x40)

MSGID_SZ = 8

TYPE_IDENTITY = 0
TYPE_PUB = 1
TYPE_RPC = 2
TYPE_ADMIN = 3
TYPE_RESPONSE = 4
TYPE_ACK = 5
TYPE_ACK2 = 6
TYPE_UNREGISTER = 9
TYPE_REGISTER = 10
TYPE_ATTESTATION = 11

STS_OK = 0
STS_ERROR = 0x0A
STS_CHALLENGE = 0x0B            # 11
STS_IDENTIFICATION_ERROR = 0X14  # 20
STS_METHOD_EXCEPTION = 0X28     # 40
STS_METHOD_ARGS_ERROR = 0X29    # 41
STS_METHOD_NOT_FOUND = 0X2A     # 42
STS_METHOD_UNAVAILABLE = 0X2B   # 43
STS_METHOD_LOOPBACK = 0X2C      # 44
STS_TARGET_NOT_FOUND = 0X2D     # 45
STS_TARGET_DOWN = 0X2E          # 46
STS_UNKNOWN_ADMIN_CMD = 0X2F    # 47
STS_NAME_ALREADY_TAKEN = 0X3C  # 60

DATAFRAME_TAG = 80

retcode = {
    STS_OK: 'OK',
    STS_ERROR: 'internal error',
    STS_IDENTIFICATION_ERROR: 'IDENTIFICATION_ERROR',
    STS_METHOD_EXCEPTION: 'METHOD_EXCEPTION',
    STS_METHOD_ARGS_ERROR: 'METHOD_ARGS_ERROR',
    STS_METHOD_NOT_FOUND: 'METHOD_NOT_FOUND',
    STS_METHOD_UNAVAILABLE: 'METHOD_UNAVAILABLE',
    STS_METHOD_LOOPBACK: 'METHOD_LOOPBACK',
    STS_TARGET_NOT_FOUND: 'TARGET_NOT_FOUND',
    STS_TARGET_DOWN: 'TARGET_DOWN',
    STS_UNKNOWN_ADMIN_CMD: 'UNKNOWN_ADMIN_CMD',
    STS_NAME_ALREADY_TAKEN: 'NAME_ALREADY_TAKEN',
}

BROKER_CONFIG = '__config__'
COMMAND = 'cmd'
ADD_INTEREST = 'subscribe'
REMOVE_INTEREST = 'unsubscribe'
ADD_IMPL = 'expose'
REMOVE_IMPL = 'unexpose'


def msgid():
    """Return an array of MSGID_SZ random bytes."""
    return int.from_bytes(os.urandom(MSGID_SZ))


def msgid_slot(slot: int):
    """Return MSGID_SZ-4 random bytes and a 4 bytes slot embedded value."""
    mid = int.from_bytes(os.urandom(MSGID_SZ))
    return mid & 0xFFFFFFFF00000000 | (slot & 0xFFFFFFFF)


def bytes2id(byte_data: bytearray) -> int:
    """
    Converts a MSGID_SZ-byte bytearray to a UInt128 integer.

    Args:
        byte_data: A bytearray of MSGID_SZ bytes.

    Returns:
        A Python integer representing the UInt128 value.
    """
    if len(byte_data) != MSGID_SZ:
        raise ValueError(f"bytearray must be exactly {MSGID_SZ} bytes long")

    # Convert bytes to integer (assuming little-endian, adjust if big-endian)
    return from_bytes(byte_data)


class RembusException(Exception):
    """Base class for all Rembus exceptions."""


class RembusTimeout(RembusException):
    """Raised when a Rembus message that expects a response times out."""

    def __str__(self):
        return 'request timeout'


class RembusConnectionClosed(RembusException):
    """Raised when a Rembus connection is closed unexpectedly."""

    def __str__(self):
        return 'connection down'


class RembusError(RembusException):
    """Raised when a Rembus message returns a generic error."""

    def __init__(self, status_code: int, msg: str | None = None):
        self.status = status_code
        self.message = msg

    def __str__(self):
        if self.message:
            return f'{retcode[self.status]}:{self.message}'
        else:
            return f'{retcode[self.status]}'


def to_bytes(val: int) -> bytes:
    """Convert an int to MSGID_SZ bytes"""
    return val.to_bytes(MSGID_SZ, 'little')


def from_bytes(bval: bytes | bytearray) -> int:
    """Convert MSGID_SZ bytes to an int"""
    return int.from_bytes(bval, 'little')


class RembusMsg(BaseModel):
    """Rembus message"""
    _twin: Any = PrivateAttr(default=None)

    @property
    def twin(self):
        """Return the Twin that owns the message"""
        return self._twin

    @twin.setter
    def twin(self, rb):
        self._twin = rb

    def to_payload(self, enc: int) -> bytes | str:
        """Return the message list of values to encode"""
        raise RuntimeError("abstract rembus message")


class RpcReqMsg(RembusMsg):
    """RPC request packet."""
    id: int
    topic: str
    data: Any = None
    target: str | None = None

    def to_payload(self, enc: int) -> bytes | str:
        """Return the RpcReqMsg list of values to encode"""
        if enc == CBOR:
            return cbor2.dumps(
                [TYPE_RPC, to_bytes(self.id), self.topic,
                 self.target, self.data]
            )

        return json.dumps({
            "jsonrpc": "2.0",
            "id": self.id,
            "method": self.topic,
            "params": self.data
        })


class ResMsg(RembusMsg):
    """Response packet."""
    id: int
    status: int
    data: Any = None
    _reqdata: Any = PrivateAttr()

    def to_payload(self, enc: int) -> bytes | str:
        """Return the ResMsg list of values to encode"""
        if enc == CBOR:
            return cbor2.dumps([
                TYPE_RESPONSE,
                to_bytes(self.id),
                self.status,
                df2tag(self.data)
            ])

        if self.status == STS_OK or self.status == STS_CHALLENGE:
            restype = "result"
        else:
            restype = "error"

        return json.dumps({
            "jsonrpc": "2.0",
            "id": self.id,
            restype: {
                "type": TYPE_RESPONSE,
                "sts": self.status,
                "data": self.data
            }
        })


class PubSubMsg(RembusMsg):
    """Pub/Sub message packet."""
    id: int | None = None
    topic: str
    data: Any = None
    flags: UInt8 = QOS0  # type: ignore[valid-type]
    slot: int | None = None

    def to_payload(self, enc: int) -> bytes | str:
        """Return the PubSubMsg list of values to encode"""

        if self.id is not None:
            if self.slot is not None:
                self.flags |= SLOT_FLAG

            if enc == CBOR:
                return cbor2.dumps([
                    TYPE_PUB | self.flags,
                    to_bytes(self.id),
                    self.topic,
                    df2tag(self.data)
                ])

            return json.dumps({
                "jsonrpc": "2.0",
                "id": self.id,
                "method": self.topic,
                "params": {
                    "type": self.flags,
                    "data": self.data
                }
            })
        else:
            if self.slot is None:
                if enc == CBOR:
                    return cbor2.dumps(
                        [TYPE_PUB, self.topic, df2tag(self.data)]
                    )

                return json.dumps({
                    "jsonrpc": "2.0",
                    "method": self.topic,
                    "params": self.data
                })
            else:
                self.flags |= SLOT_FLAG
                mid = to_bytes(self.slot)
                if enc == CBOR:
                    return cbor2.dumps([
                        TYPE_PUB | self.flags,
                        mid,
                        self.topic,
                        df2tag(self.data)
                    ])

                return json.dumps({
                    "jsonrpc": "2.0",
                    "id": self.slot,
                    "method": self.topic,
                    "params": self.data
                })


class AckMsg(RembusMsg):
    """Pub/Sub Ack message for QOS1 and QOS2"""
    id: int

    def to_payload(self, enc: int) -> bytes | str:
        """Return the AckMsg list of values to encode"""
        if enc == CBOR:
            return cbor2.dumps([TYPE_ACK, to_bytes(self.id)])

        return json.dumps({
            "jsonrpc": "2.0",
            "id": self.id,
            "result": {
                "type": TYPE_ACK
            }
        })


class Ack2Msg(RembusMsg):
    """Pub/Sub Ack2 message for QOS2"""
    id: int

    def to_payload(self, enc: int) -> bytes | str:
        """Return the Ack2Msg list of values to encode"""
        if enc == CBOR:
            return cbor2.dumps([TYPE_ACK2, to_bytes(self.id)])

        return json.dumps({
            "jsonrpc": "2.0",
            "id": self.id,
            "result": {
                "type": TYPE_ACK2
            }
        })


class AdminMsg(RembusMsg):
    """AdminMsg packet."""
    id: int
    topic: str
    data: Any = None

    def __str__(self):
        return f'AdminMsg:{self.topic}: {self.data}'

    def to_payload(self, enc: int) -> bytes | str:
        """Return the AdminMsg list of values to encode"""
        if enc == CBOR:
            return cbor2.dumps(
                [TYPE_ADMIN, to_bytes(self.id), self.topic, self.data]
            )

        return json.dumps({
            "jsonrpc": "2.0",
            "id": self.id,
            "method": self.topic,
            "params": {
                "type": TYPE_ADMIN,
                "data": self.data
            }
        })


class IdentityMsg(RembusMsg):
    """IdentityMsg packet.
    This message is sent by the component to identify itself
    to the remote peer.
    """
    id: int
    cid: str

    def __str__(self):
        return f'IdentityMsg:{self.cid}'

    def to_payload(self, enc: int) -> bytes | str:
        """Return the IdentityMsg payload"""
        if enc == CBOR:
            return cbor2.dumps([TYPE_IDENTITY, to_bytes(self.id), self.cid])

        return json.dumps({
            "jsonrpc": "2.0",
            "id": self.id,
            "method": "__identity__",
            "params": {
                "type": TYPE_IDENTITY,
                "cid": self.cid
            }
        })


class AttestationMsg(RembusMsg):
    """AttestationMsg packet.
    This message is sent by the component to authenticate
    its identity to the remote peer.
    """
    id: int
    cid: str
    signature: bytes | str

    def __str__(self):
        return f'AttestationMsg:{self.cid}'

    def to_payload(self, enc: int) -> bytes | str:
        """Return the AttestationMsg list of values to encode"""
        if enc == CBOR:
            return cbor2.dumps(
                [TYPE_ATTESTATION, to_bytes(self.id), self.cid, self.signature]
            )

        return json.dumps({
            "jsonrpc": "2.0",
            "id": self.id,
            "method": self.cid,
            "params": {
                "type": TYPE_ATTESTATION,
                "signature": self.signature
            }
        })


class RegisterMsg(RembusMsg):
    """RegisterMsg packet.
    This message is sent by the component to register
    its public key to the remote peer.
    """
    id: int
    cid: str
    pin: str
    pubkey: bytes | str
    type: int

    def to_payload(self, enc: int) -> bytes | str:
        """Return the RegisterMsg list of values to encode"""
        if enc == CBOR:
            return cbor2.dumps([
                TYPE_REGISTER,
                to_bytes(self.id),
                self.cid,
                self.pin,
                self.pubkey,
                self.type
            ])

        return json.dumps({
            "jsonrpc": "2.0",
            "id": self.id,
            "method": self.cid,
            "params": {
                "type": TYPE_REGISTER,
                "pin": self.pin,
                "key_val": self.pubkey,
                "key_type": self.type
            }
        })


class UnregisterMsg(RembusMsg):
    """UnregisterMsg packet.
    This message is sent by the component to unregister
    its public key from the remote peer.
    """
    id: int

    def to_payload(self, enc: int) -> bytes | str:
        """Return the UnregisterMsg list of values to encode"""
        if enc == CBOR:
            return cbor2.dumps([TYPE_UNREGISTER, to_bytes(self.id)])

        return json.dumps({
            "jsonrpc": "2.0",
            "id": self.id,
            "method": "__unregister__",
            "params": {
                "type": TYPE_UNREGISTER
            }
        })


def isregistered(router_id, rid: str):
    """Check if a component identified by `rid` is registered in the router."""
    try:
        rs.key_file(router_id, rid)
        return True
    except FileNotFoundError:
        return False


def tohex(bs: bytes):
    """Return a string with bytes as hex numbers with 0xNN format."""
    return ' '.join(f'0x{x:02x}' for x in bs)


def jsonprc_request(pkt, msg_id, params) -> RembusMsg:
    """Parse a JSON_RPC request"""
    if isinstance(params, list):
        # Default to RPC request
        return RpcReqMsg(id=msg_id, topic=pkt["method"], data=params)
    else:
        msg_type = params.get("type")
        if msg_type in [QOS1, QOS2]:
            return PubSubMsg(
                id=msg_id,
                topic=pkt["method"],
                data=pkt.get("data"),
                flags=msg_type
            )
        elif msg_type == TYPE_IDENTITY:
            return IdentityMsg(id=msg_id, cid=params["cid"])
        elif msg_type == TYPE_ADMIN:
            return AdminMsg(
                id=msg_id,
                topic=pkt["method"],
                data=pkt.get("data")
            )
        elif msg_type == TYPE_ATTESTATION:
            sig = params.get("signature")
            return AttestationMsg(id=msg_id, cid=pkt["method"], signature=sig)
        elif msg_type == TYPE_REGISTER:
            pubkey = params.get("key_val")
            if isinstance(pubkey, str):
                pubkey = base64.b64decode(pubkey)

            return RegisterMsg(
                id=msg_id,
                cid=pkt["method"],
                pin=params.get("pin"),
                pubkey=pubkey,
                type=params.get("key_type")
            )
        elif msg_type == TYPE_UNREGISTER:
            return UnregisterMsg(
                id=msg_id
            )

        raise ValueError(f"{pkt}:invalid JSON-RPC request")


def jsonprc_response(pkt, msg_id, result) -> RembusMsg:
    """Parse a JSON_RPC success response"""

    msg_type = result.get("type")
    if msg_type == TYPE_RESPONSE or msg_type is None:
        status = result.get("sts", STS_OK)
        return ResMsg(id=msg_id, status=status, data=result.get("data"))
    elif msg_type == TYPE_ACK:
        return AckMsg(id=msg_id)
    elif msg_type == TYPE_ACK2:
        return Ack2Msg(id=msg_id)

    raise ValueError(f"{pkt}:invalid JSON-RPC response")


def jsonrpc_parse(payload) -> RembusMsg:
    """Get a Rembus message from a JSON-RPC string payload."""
    pkt = json.loads(payload)
    msg_id = pkt.get("id")
    if msg_id:
        # request-response message
        params = pkt.get("params")
        if params is not None:
            return jsonprc_request(pkt, msg_id, params)

        result = pkt.get("result")
        if result:
            return jsonprc_response(pkt, msg_id, result)

        err = pkt.get("error")
        if err:
            return jsonprc_response(pkt, msg_id, err)

    else:
        return PubSubMsg(topic=pkt["method"], data=pkt.get("data"))

    raise ValueError(f"{pkt}:JSON-RPC invalid payload")


def cbor_parse(pkt) -> RembusMsg:
    """Parse a CBOR packet into a Rembus message."""
    type_byte = pkt[0]
    mtype = type_byte & 0x0F
    flags = type_byte & 0xF0

    def f(idx):  # shortcut for from_bytes
        return from_bytes(pkt[idx])

    # Special cases first (because of flag-dependent structure)
    if mtype == TYPE_PUB:
        if flags == QOS0:
            return PubSubMsg(topic=pkt[1], data=pkt[2])
        return PubSubMsg(id=f(1), topic=pkt[2], data=pkt[3], flags=flags)

    if mtype == TYPE_RESPONSE:
        return ResMsg(
            id=f(1), status=pkt[2], data=pkt[3] if len(pkt) > 3 else None
        )

    # Dispatch table for the rest
    table = {
        TYPE_RPC:      lambda: RpcReqMsg(id=f(1),
                                         topic=pkt[2],
                                         target=pkt[3],
                                         data=pkt[4]),
        TYPE_ACK:      lambda: AckMsg(id=f(1)),
        TYPE_ACK2:     lambda: Ack2Msg(id=f(1)),
        TYPE_ADMIN:    lambda: AdminMsg(id=f(1), topic=pkt[2], data=pkt[3]),
        TYPE_IDENTITY: lambda: IdentityMsg(id=f(1), cid=pkt[2]),
        TYPE_ATTESTATION: lambda: AttestationMsg(id=f(1),
                                                 cid=pkt[2],
                                                 signature=pkt[3]),
        TYPE_REGISTER: lambda: RegisterMsg(id=f(1),
                                           cid=pkt[2],
                                           pin=pkt[3],
                                           pubkey=pkt[4],
                                           type=pkt[5]),
        TYPE_UNREGISTER: lambda: UnregisterMsg(id=f(1)),
    }

    try:
        return table[mtype]()
    except KeyError as e:
        raise ValueError(f"unknown message type: {mtype}") from e


def decode_dataframe(data: bytes) -> IntoFrame:
    """Decode a CBOR tagged value `data` to a pandas dataframe."""
    writer = pa.BufferOutputStream()
    writer.write(data)
    buf: pa.Buffer = writer.getvalue()
    with pa.ipc.open_stream(buf) as reader:
        table = reader.read_all()

    return pl.from_arrow(table)


def encode_dataframe(df: IntoFrame) -> cbor2.CBORTag:
    """Encode a pandas dataframe `df` to a CBOR tag value."""
    if isinstance(df, pl.DataFrame):
        table = df.to_arrow()
    else:
        table = pa.Table.from_pandas(df)

    sink = pa.BufferOutputStream()
    with pa.ipc.new_stream(sink, table.schema) as writer:
        writer.write(table)
    buf = sink.getvalue()
    stream = pa.input_stream(buf)
    return cbor2.CBORTag(DATAFRAME_TAG, stream.read())


def encode(msg: list[Any]) -> bytes:
    """Encode message `msg`."""
    return cbor2.dumps(msg)


def tag2df(data: Any) -> Any:
    """Loop over `data` items and decode tagged values to dataframes."""
    if isinstance(data, list):
        for idx, val in enumerate(data):
            if isinstance(val, cbor2.CBORTag) and val.tag == DATAFRAME_TAG:
                data[idx] = decode_dataframe(val.value)
    elif isinstance(data, cbor2.CBORTag):
        return decode_dataframe(data.value)
    return data


def df2tag(data: Any) -> Any:
    """Loop over `data` items and encode dataframes to tag values."""
    if isinstance(data, tuple):
        lst: List[Any] = []
        for idx, val in enumerate(data):
            if isinstance(val, pl.DataFrame) or isinstance(val, pd.DataFrame):
                lst.append(encode_dataframe(val))
            else:
                lst.append(val)
        return lst
    elif isinstance(data, list):
        for idx, val in enumerate(data):
            if isinstance(val, pl.DataFrame) or isinstance(val, pd.DataFrame):
                data[idx] = encode_dataframe(val)

    elif isinstance(data, pl.DataFrame) or isinstance(data, pd.DataFrame):
        data = encode_dataframe(data)
    return data


def rsa_private_key():
    """Generate a new RSA private key."""
    return rsa.generate_private_key(public_exponent=65537, key_size=2048)


def ecdsa_private_key():
    """Generate a new ECDSA private key using SECP256R1 curve."""
    return ec.generate_private_key(ec.SECP256R1(), default_backend())


def pem_public_key(private_key: PrivateKeyTypes) -> bytes:
    """Return the public key in PEM format from a private key."""
    return private_key.public_key().public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )


def save_private_key(cid: str, private_key: PrivateKeyTypes):
    """Save the private key to a file in the rembus directory."""
    kdir = os.path.join(rs.rembus_dir(), cid)

    if not os.path.exists(kdir):
        os.makedirs(kdir)

    fn = os.path.join(kdir, ".secret")
    private_key_file = open(fn, "wb")

    pem_private_key = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption()
    )

    private_key_file.write(pem_private_key)
    private_key_file.close()


def load_private_key(cid: str) -> PrivateKeyTypes:
    """Load the private key from a file in the rembus directory."""
    fn = os.path.join(rs.rembus_dir(), cid, ".secret")
    # fn = os.path.join(rembus_dir(), cid)
    with open(fn, "rb") as key_file:
        private_key = serialization.load_pem_private_key(
            key_file.read(), password=None)

    return private_key


def load_public_key(router, cid: str):
    """Load the public key from a file in the rembus directory."""
    fn = rs.key_file(router.id, cid)
    try:
        with open(fn, "rb") as f:
            public_key = serialization.load_pem_public_key(
                f.read(),
            )
    except ValueError:
        try:
            with open(fn, "rb") as f:
                public_key = serialization.load_der_public_key(
                    f.read(),
                )
        except ValueError as e:
            raise ValueError(
                f"Could not load public key from file: {fn}") from e

    return public_key


def save_pubkey(router_id: str, cid: str, pubkey: bytes, keytype: int):
    """Save the public key to a file in the rembus directory."""
    name = rs.key_base(router_id, cid)
    keyformat = "der"
    # check if pubkey start with -----BEGIN chars
    if pubkey[0:10] == bytes(
        [0x2d, 0x2d, 0x2d, 0x2d, 0x2d, 0x42, 0x45, 0x47, 0x49, 0x4e]
    ):
        keyformat = "pem"

    if keytype == SIG_RSA:
        typestr = "rsa"
    else:
        typestr = "ecdsa"

    fn = f"{name}.{typestr}.{keyformat}"
    with open(fn, "wb") as f:
        f.write(pubkey)


def remove_pubkey(router, cid: str):
    """Remove the public key file for a component."""
    fn = rs.key_file(router.id, cid)
    os.remove(fn)



"""Synchronous APIs."""
import asyncio
import logging
import threading
from types import TracebackType
from typing import Any, Callable, Coroutine, Optional, Type, List
from rembus import (
    component,
    RbURL,
    CBOR,
    SIG_RSA,
)

from rembus.protocol import RembusConnectionClosed

logger = logging.getLogger(__name__)


class AsyncLoopRunner:
    """:meta private:"""

    def __init__(self):
        self.loop = asyncio.new_event_loop()
        self._thread = threading.Thread(target=self._start_loop, daemon=True)
        self._thread.start()

    def _start_loop(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    def run(self, coro: Coroutine[Any, Any, Any]) -> Any:
        """Run a coroutine in the event loop and return its result."""
        future = asyncio.run_coroutine_threadsafe(coro, self.loop)
        return future.result()

    def shutdown(self):
        """Shutdown the event loop and wait for the thread to finish."""
        if not self.loop.is_closed():
            self.loop.call_soon_threadsafe(self.loop.stop)

            # Wait briefly for the thread to exit
            self._thread.join(timeout=1.0)

            # Cancel any pending tasks to avoid dangling transports
            pending = asyncio.all_tasks(self.loop)
            for task in pending:
                task.cancel()
            if pending:
                self.loop.run_until_complete(
                    asyncio.gather(*pending, return_exceptions=True))


class node:  # pylint: disable=invalid-name
    """The synchronous Rembus twin."""

    def __init__(
        self,
        url: str | List[str] | None = None,
        name: str | None = None,
        port: int | None = None,
        secure: bool = False,
        enc: int = CBOR
    ):
        self._runner = AsyncLoopRunner()
        self._rb = self._runner.run(component(url, name, port, secure, enc))

    @property
    def router(self):
        """The router object associated with this node."""
        return self._rb.router

    @property
    def uid(self):
        """The RbUrl object for this node."""
        return self._rb.uid

    @property
    def rid(self):
        """The unique Rembus id for this node."""
        return self._rb.rid

    def isopen(self) -> bool:
        """Check if the connection is open."""
        return self._rb.isopen()

    def isrepl(self) -> bool:
        """Check if the connection is a REPL (Read-Eval-Print Loop)."""
        return self._rb.isrepl()

    def inject(self, ctx: Any):
        """Initialize the context object."""
        return self._rb.inject(ctx)

    def exec(self, fn, *args, **kwargs):
        """Execute a method in the event loop and return its result."""
        if self._runner is not None:
            coro = fn(*args, **kwargs)
            return self._runner.run(coro)

        raise RembusConnectionClosed()

    def register(self, rid: str, pin: str, scheme: int = SIG_RSA):
        """Register a component with the given rid."""
        return self.exec(self._rb.register, rid, pin, scheme)

    def unregister(self):
        """Unregister the component."""
        return self.exec(self._rb.unregister)

    def direct(self, target: str, topic: str, *args: Any):
        """Send a direct RPC request to the target component."""
        return self.exec(self._rb.direct, target, topic, *args)

    def rpc(self, topic: str, *args: Any):
        """Send a RPC request."""
        return self.exec(self._rb.rpc, topic, *args)

    def publish(self, topic: str, *args: Any, **kwargs):
        """Publish a message to the specified topic."""
        return self.exec(self._rb.publish, topic, *args, **kwargs)

    def put(self, topic: str, *args: Any, **kwargs):
        """Publish a message to the specified topic."""
        return self.exec(
            self._rb.publish, self.rid + '/' + topic, *args, **kwargs
        )

    def subscribe(self,
                  fn: Callable[..., Any],
                  retroactive: bool = False,
                  topic: Optional[str] = None):
        """
        Subscribe the function to the corresponding topic.
        """
        return self.exec(self._rb.subscribe, fn, retroactive, topic)

    def unsubscribe(self, fn: Callable[..., Any]):
        """
        Unsubscribe the function from the corresponding topic.
        """
        return self.exec(self._rb.unsubscribe, fn)

    def expose(self, fn: Callable[..., Any]):
        """
        Expose the function as a remote procedure call(RPC) handler.
        """
        return self.exec(self._rb.expose, fn)

    def unexpose(self, fn: Callable[..., Any]):
        """
        Unexpose the function as a remote procedure call(RPC) handler.
        """
        return self.exec(self._rb.unexpose, fn)

    def reactive(self):
        """
        Set the component to receive published messages on subscribed topics.
        """
        return self.exec(self._rb.reactive)

    def unreactive(self):
        """
        Set the component to stop receiving published
        messages on subscribed topics.
        """
        return self.exec(self._rb.unreactive)

    def wait(self, timeout: float | None = None):
        """
        Start the twin event loop that wait for rembus messages.
        """
        return self.exec(self._rb.wait, timeout)

    def close(self):
        """Close the connection and clean up resources."""
        self.exec(self._rb.close)
        self._runner.shutdown()
        self._runner = None

    def __enter__(self):
        return self

    def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]) -> Optional[bool]:
        self.close()


def register(rid: str, pin: str, scheme: int = SIG_RSA, enc: int = CBOR):
    """Provisions the component with rid identifier."""
    rburl = RbURL(rid)
    rb = node(rid, enc=enc)
    try:
        rb.register(rburl.id, pin, scheme)
    except Exception as e:
        logger.error("[%s] register: %s", rid, e)
        raise
    finally:
        rb.close()



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("rembus")
except PackageNotFoundError:
    __version__ = "0.0.0"

# import logging
from .core import component, RbURL  # noqa: F401
from .settings import (  # noqa: F401
    rembus_dir,
    DEFAULT_BROKER,
    TENANTS_FILE
)
from .protocol import (  # noqa: F401
    CBOR,
    JSON,
    QOS0,
    QOS1,
    QOS2,
    SIG_ECDSA,
    SIG_RSA
)
from .sync import node, register

__all__ = [
    'component',
    'node',
    'register',
]
